#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Use provided GetOpt if compiling on Windows. */
#ifdef WIN32
	#include "getopt.h"
#else
	#include <getopt.h>
#endif


#define VERSION "2.0"
/*
   amicon: an Amiga icon to ppm/xpm convertor.
   Copyright 2001 Adam Sampson <azz@gnu.org>

   Modified for Windows by Gary Harris 27/10/03.

   amicon is free software; you can redistribute and/or modify it
   under the terms of that license as published by the Free Software
   Foundation; either version 2 of the License, or (at your option)
   any later version.
   
   amicon is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with amicon; see the file COPYING. If not, write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA, or see http://www.gnu.org/.
*/



typedef unsigned char u8;
typedef signed char s8;
typedef unsigned short u16;
typedef unsigned int u32;
u16 magic, version, type;
u32 image1, image2, tool, num_tooltypes;
int xpm = 0, transparent = -1;
u8 mwb_palette[] = {
	149, 149, 149,
	0, 0, 0,
	255, 255, 255,
	59, 103, 162,
	123, 123, 123,
	175, 175, 175,
	170, 144, 124,
	255, 169, 151
};

#ifdef WIN32
	FILE	*file;
#endif

void die(char *msg) {

#ifdef WIN32
	printf("%s\n", msg);
#else
	fprintf(stderr, "%s\n", msg);
#endif
		exit(1);
}

u16 get16(u8 *p) {
	return (*p * 256) + *(p+1);
}

u32 get32(u8 *p) {
	return ((*p * 256 + *(p+1)) * 256 + *(p+2)) * 256 + *(p+3);
}

int read8(u8 *buf, int len) {
	
#ifdef WIN32
	len = fread(buf, 1, len, file);
#else
	len = fread(buf, 1, len, stdin);
#endif

	return /*fread(buf, 1, len, stdin) == */len;
}

void read_header() {
	u8 header[0x4e];

	if (!read8(header, sizeof header))
		die("unable to read header");
	magic = get16(header);
	if (magic != 0xE310)
		die("wrong magic");
	version = get16(header + 0x02);
	if (version != 1)
		die("wrong version");
	type = get16(header + 0x30);
	image1 = get32(header + 0x16);
	if (!image1)
		die("could not find image");
	image2 = get32(header + 0x1A);
	tool = get32(header + 0x32);
	if (get32(header + 0x42)) {
		u8 drawer[0x38];
		if (!read8(drawer, sizeof drawer))
			die("unable to read drawer");
	}
}

void write_ppm(u8 *bitmap, int width, int height, int ncols, u8 *palette) {
	int x, y;
	printf("P6 %d %d 255\n", width, height);
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			u8 c = 3 * bitmap[x + y*width];
			putchar(palette[c]);
			putchar(palette[c+1]);
			putchar(palette[c+2]);
		}
	}
}

void write_xpm(u8 *bitmap, int width, int height, int ncols, u8 *palette) {
	int x, y;
	printf("/* XPM */\nstatic char * amigaicon_xpm[] = {\n"
		"\"%d %d %d 1\",\n", width, height, ncols);
	for (x = 0; x < ncols; x++)
		if (x == transparent)
			printf("\"%c\tc None\",\n", 'A' + x);
		else
			printf("\"%c\tc #%02x%02x%02x\",\n",
				'A' + x,
				palette[x*3], palette[x*3+1], palette[x*3+2]);
	for (y = 0; y < height; y++) {
		putchar('"');
		for (x = 0; x < width; x++)
			putchar('A' + bitmap[x + y*width]);
		if (y == height - 1)
			printf("\"};");
		else
			printf("\",\n");
	}
}

void write_image(u8 *bitmap, int width, int height, int ncols, u8 *palette) {
	if (xpm)
		write_xpm(bitmap, width, height, ncols, palette);
	else
		write_ppm(bitmap, width, height, ncols, palette);
}

void read_icon(int print) {
	u8 header[0x14];
	u16 width, height, depth;
	u8 plane_pick, plane_onoff;
	u8 *row, *bitmap;
	int rwidth;
	int ib;

	if (!read8(header, sizeof header))
		die("unable to read icon header");
	width = get16(header + 0x04);
	height = get16(header + 0x06);
	depth = get16(header + 0x08);
	plane_pick = *(header + 0x0e);
	if (!plane_pick)
		die("icon does not have any bitplanes");
	plane_onoff = *(header + 0x0f);
	rwidth = ((width + 15) / 16) * 2;
	row = malloc(rwidth);
	bitmap = malloc(width * height);
	memset(bitmap, 0, width * height);
	for (ib = 0; ib < depth; ib++) {
		int ix, iy;
		int oslib = 1 << ib;
		if (plane_pick & oslib) {
			for (iy = 0; iy < height; iy++) {
				if (!read8(row, rwidth))
					die("unable to read row");
				for (ix = 0; ix < width; ix++)
					if (row[ix/8] & (0x80 >> (ix & 0x7)))
						bitmap[ix + iy*width] |= oslib;
			}
		} else if (plane_onoff & oslib) {
			for (iy = 0; iy < height; iy++)
				for (ix = 0; ix < width; ix++)
					bitmap[ix + iy*width] |= oslib;
		}
	}
	if (print)
		write_image(bitmap, width, height, 8, mwb_palette);
	free(bitmap);
	free(row);
}

char *read_string() {
	u8 *buf;
	u8 lenbuf[4];
	u32 len;

	if (!read8(lenbuf, sizeof lenbuf))
		die("unable to read string length");
	len = get32(lenbuf);
	buf = malloc(len + 1);
	buf[len] = '\0';
	if (len > 0)
		if (!read8(buf, len))
			die("unable to read string");
	return buf;
}

void start_tooltypes() {
	u8 countbuf[4];

	if (!read8(countbuf, sizeof countbuf))
		die("unable to read number of tooltypes");
	num_tooltypes = get32(countbuf)/4 - 1;
}

u8 *get_tooltype(char *name) {
	while (num_tooltypes > 0) {
		u8 *s = read_string();
		num_tooltypes--;
		if (strncmp(s, name, strlen(name)) == 0) {
			return s;
		}
		free(s);
	}
	return NULL;
}

u8 *decompress_newicon(u8 *src, char *name, int len, int bits) {
	u8 *s = src;
	u8 *data = malloc(len), *p = data;
	int bitmask = (1 << bits) - 1;
	int workbits, currentworkbit;

	readblock:
	workbits = 0;
	currentworkbit = 7;

	while (1) {
		int runlength = 0;
		s8 c = (s8) *s++;
		if (c == 0) {
			free(src);
			src = get_tooltype(name);
			if (!src)
				die("ran out of tooltype data");
			s = src + 4;
			goto readblock;
		}
		c -= 0x20;
		if (c <= 0) {
			c -= 0x31;
			if (c <= 0) {
				runlength = c + 0x80;
				c = 0;
			}
		}
		workbits |= c;
		while (runlength-- >= 0) {
			while (currentworkbit - bits >= 0) {
				currentworkbit -= bits;
				*p = (workbits >> currentworkbit) & bitmask;
				p++;
				len--;
				if (len == 0) {
					free(src);
					return data;
				}
			}
			workbits <<= 7;
			currentworkbit += 7;
		}
	}
}

void read_newicon(char *name) {
	u8 *s, *palette, *bitmap;
	u8 transparent, width, height;
	u16 colors;
	int bits;

	s = get_tooltype(name);
	if (!s)
		die("newicon not found");
	if (strlen(s) < 4)
		die("newicon header too short");
	transparent = s[4] - 0x21;
	width = s[5] - 0x21;
	height = s[6] - 0x21;
	colors = ((s[7] - 0x21) << 6) | (s[8] - 0x21);
	if (width > 95 || height > 95)
		die("illegal newicon size");
	if (colors > 256)
		die("illegal newicon palette size");

	palette = decompress_newicon(strdup(s + 9), name, colors * 3, 8);
	free(s);

	bits = 1;
	while ((colors - 1) >> bits)
		bits++;
	bitmap = decompress_newicon(strdup(""), name, width * height, bits);

	write_image(bitmap, width, height, colors, palette);

	free(palette);
	free(bitmap);
}

int print_help(void)
{
#ifdef WIN32
	printf("amicon " VERSION " by Adam Sampson <azz@gnu.org>.\n"
	"Modified for Windows by Gary Harris <raider@pacific.net.au> 27/10/03.\n\n"
#else
	fprintf(stderr, "amicon " VERSION " by Adam Sampson <azz@gnu.org>\n\n"
	"Modified for Windows by Gary Harris <raider@pacific.net.au> 27/10/03.\n\n"
#endif
		"-u      Convert old unselected icon (default)\n"
		"-s      Convert old selected icon\n"
		"-U      Convert unselected NewIcon\n"
		"-S      Convert selected NewIcon\n"
		"-t n    Make colour number n transparent (usually 0)\n"
		"-p      Generate PPM output (default)\n"
		"-x      Generate XPM output\n"
		"-h      This help\n");
	exit(0);
}


typedef enum {
	OLD1, OLD2, NEW1, NEW2
} mode;

int main(int argc, char **argv) {
	mode mode = OLD1;

	/* Check for file name only and print help.*/
	if(argc == 1)
		print_help();

	while (1) {

#ifdef WIN32
		char* pszParam;
        char c = GetOption(argc, argv, "usUSt:pxh", &pszParam);
		if (c < 1)
			c = pszParam[1];
//			break;

#else		
		char c = getopt(argc, argv, "usUSt:pxh");
		if (c == -1)
			break;
#endif

		switch (c) {
		case 'u':
			mode = OLD1;
			break;
		case 's':
			mode = OLD2;
			break;
		case 'U':
			mode = NEW1;
			break;
		case 'S':
			mode = NEW2;
			break;
		case 't':
#ifdef WIN32
			if(!pszParam)
				die("Faulty transparency argument.\nMake sure that a number appears after \"-t\" on the command line.");
			transparent = atoi(pszParam);
#else		
			transparent = atoi(optarg);
#endif
			break;
		case 'p':
			xpm = 0;
			break;
		case 'x':
			xpm = 1;
			break;
		case 'h':
			print_help();
		default:{
			char buf[18];
			sprintf(buf, "%s%c", "Unknown option: ", c);
			die(buf);
			}
		}
	}

#ifdef WIN32
  /* Open as a file under Windows. */
	if((file = fopen(argv[argc - 1], "rb")) == NULL)
		die("Couldn't open icon file.\nMake sure that the icon name appears last on the command line.");
#endif

	read_header();
	read_icon(mode == OLD1);
	if (image2)
		read_icon(mode == OLD2);
	if (tool) 
		free(read_string());
	start_tooltypes();
	if (mode == NEW1)
		read_newicon("IM1=");
	else if (mode == NEW2)
		read_newicon("IM2=");
	
	return 0;
}

